<?php
 /** 
 * $Id: carbon_calculate.inc,v 1.19 2007/03/07 14:03:59 johnackers Exp $ 
 * 
 * Drupal carbon module (calculations) 
 *
 * Created on 18-Oct-2006
 * 
 * @file 
 * The calculations used to build carbon footprint.
 * 
 * Different carbon_stamps have a different scope, some declare 
 * emissions on one date, others declare emissions over a period
 * of time. 
 *    
 * Some sources of emissions e.g. from air travel need special 
 * formulas for calculating resulting CO2 because fuel used in 
 * aircraft takeoff has to be taken into consideration. 
 */


/** 
 * Create a set of results for a given period (usually a year) 
 * based on on all the available carbon_stamps. Maintain 
 * sector (e.g. home, tranport) and grand totals.  
 * 
 * @param $ar array of all carbon_stamps (with associated carbon_source fields)
 * @param $start first date the generated carbon footprint 
 * @param $end last date of the generated carbon footprint
 *  
 * @return a complex structure that has the carbon footprint for this period
 */


function carbon_calculate_total(&$ar, $start, $end)
{  
  $results = new Result(); 
  $results->grand_kwh0 = 0 ;   $results->grand_kwh1 = 0 ;  $results->grand_co2 = 0 ;
  $results->sector_kwh0 = array() ; $results->sector_kwh1 = array() ; $results->sector_co2 = array() ;
  $results->start = $start ;
  $results->end = $end ;
  $results->items = _calculate($ar, $start, $end);  
	$grand_total = 0 ; 
   
  foreach ($results->items as $result)
  {
  	$sector = $result->select[0]->sector ; 
  	if (isset($result->kwh0))
    {
     //aacc  
      if (!isset($results->sector_kwh0[$sector]))
        $results->sector_kwh0[$sector] = 0 ;

      $results->sector_kwh0[$sector] += $result->kwh0 ;       
      $results->grand_kwh0 += $result->kwh0 ; 
    }
    if (isset($result->kwh1))
    {
     //aacc  
      if (!isset($results->sector_kwh1[$sector]))
        $results->sector_kwh1[$sector] = 0 ;

      $results->sector_kwh1[$sector] += $result->kwh1 ;       
      $results->grand_kwh1 += $result->kwh1 ; 
    }
    if (isset($result->co2))
    {
      if (!isset($results->sector_co2[$sector]))
        $results->sector_co2[$sector] = 0 ;

      $results->sector_co2[$sector] += $result->co2 ;       
      $results->grand_co2 += $result->co2 ; 
    }
    // need to execute this once.
    $results->enablekwh0 = $result->select[0]->enablekwh0 ; 
    $results->enablekwh1 = $result->select[0]->enablekwh1 ; 
    $results->enableco2 =  $result->select[0]->enableco2 ; 
  } 
  return $results ;       
}

/** 
 * Aggregate all similar carbon stamps and build an array 
 * of results of CO2 calculations.
 * 
 * @param $ar array of all carbon_stamps (with associated carbon_source fields)
 * @param $start first date the generated carbon footprint 
 * @param $end last date of the generated carbon footprint
 *  
 * @return an array of results. Each result contains text and/or value.
 */

 
function _calculate(&$ar, $start, $stop)
{  
  $dgroup = array();     
  foreach ($ar as $node)
  {
    $dgroup[] = _concat_same_value($node);
  }

  // each key represents a group of readings
  // that can be considered identical 
  
  $keys = array_unique($dgroup); 
  $results = array();              
  foreach ($keys as $subgroup)
  {
  	// pull out the first group and 
    // select all the nodes that match that group

    $select = array(); 
    $column_notes = "" ;   
    foreach ($ar as $node2)
    {
      if (_concat_same_value($node2) != $subgroup) 
        continue ; 
      $select[] = $node2;  
      $node2->total = null ;
    }
    
    $fname = '_scope_'.$select[0]->scope; 

    if (function_exists($fname))
    {
      $r = ($fname($select, $start, $stop));
//      if ($r == null)
//        drupal_set_message($fname." returned NULL value", "error");
      $r->select = $select ;   // attach the original data 
      $results[] = $r; 
    }   
  }
  return $results ; 
}

/**
 * identify stamps that are identical so that they can be
 * group together. 
 * 
 * meter readings (scope == 1) are handled as a group, so stamps 
 * with a different value should be group together. 
 * 
 */


function _concat_same_value($node)
{
  $hashcode = $node->title.'.'.$node->code.'.'.$node->scope.'.'.$node->adjustment.'.'.$node->sourceclass;
  switch ($node->scope)
  {
    case 0 : 
    case 2 :
             return $hashcode.'.'.$node->reading.'.'.$node->enddate.'.'.$node->startdate;  
    case 1 : return $hashcode ;  	
  }
}

/**
 * identify stamps that are identical that can presented in a group on a form
 * and the only the field that needs to be changed is enddate and reading.
 */

function _concat($node){
  return $node->title.'.'.$node->code.'.'.$node->scope.'.'.$node->adjustment.'.'.$node->sourceclass; 
}


/**
 * todo check that no ; or : present
 */

function _evaluate($expression)
{
  $expression = trim($expression); 
  $exp1 = substr($expression,0,1); 
  if (strcmp($exp1,"=")==0)
    $expression = substr($expression,1);

  return eval('return '.$expression.';'); 
}


class Result 
{

}    



function _sort_by_enddate($ar)
{
  $ar1 = array();
  foreach ($ar as $node)
  {
    $ar1[$node->enddate] = $node ;      
  }       
  ksort($ar1);
  return array_values($ar1);
}


/**
 * There are two scenarios.  
 * 1. the emissions occur on a single date (e.g. air travel) 
 * 2. the emissions are spread over a month , year or more. 
 */

function _scope_0(&$ar, $start, $stop)
{    
  $result = new Result(); 
  _duplicate_check($ar);   
 
  foreach ($ar as $node)  
  {  	
    if (empty($node->startdate) || $node->startdate == $node->enddate)
    {
        // this is the simple case where all the emissions 
        // were released  on the same date 
        // (take care with < and <= etc to include all stamps in 
        // at least one accounting period (a year)
        //
         
      if ($node->enddate < $start) continue ; 
      if ($node->enddate >= $stop) continue ;
      _invoke_calculate($result, $node, 1);
    }
    else
    {
      // this is the case where the emissions spread across a long 
      // period of time e.g. the carbon stamp might refer to emissions
      // from june 05 to june 06 and so would appear in two separate years. 
      
      // explanation of below: 
	    // Let's say user inputs carbon stamp of 1000 Kg of CO2 over 10 years. 
      // We want to allocate the 1000Kg over the ten years as evenly as possible.
	    // So we calculate an adjustment which equals 
	    //    number of emitting days inside our period of interest (i.e between $start and $end) / 
	    //    total number of emitting days. 

      $begin = max($node->startdate, $start);
	    $end = min($node->enddate, $stop); 
	    if ($end > $begin)
      {	  
	      $adjust_for_period = ($end - $begin) / ($node->enddate - $node->startdate);   
	      _invoke_calculate($result, $node, $adjust_for_period);
	    }
    }  
    return $result ; // process only 1 stamp, ignore duplicates
	}
}

/**
 * meter readings 
 */


function _scope_1(&$ar1, $start, $end)
{ 
  $result = new Result(); 
  $result->error = array();
    
  $num = sizeof($ar1);

  if ($num == 0)
  {
    $result->error = "No readings" ; 
    return $result ; 
  }

  //  drupal_set_message("array has1 ".$num, "info"); 
  $time_sorted_map = _sort_by_enddate($ar1);
  //  drupal_set_message("array has2 ".$num, "info");

  // do not attempt to make sense of requests for  
  // consumption dates that fall entirely below or entirely 
  // above the range of available meter readings 
  
  $time_sorted_ar = array_values($time_sorted_map); 
   
  if ($end <= $time_sorted_ar[0]->enddate)
    return $result ; 
    
  if ($start > $time_sorted_ar[count($time_sorted_ar)-1]->enddate)
    return $result ; 

  $result->start = _find_closest($time_sorted_map, $start);            
  $result->end = _find_closest($time_sorted_map, $end);    


  if ($result->start == $result->end)
  {
    $result->error = "Can only find one suitable meter reading" ;
    return $result ; 
  }

  $consumption = $result->end->reading - $result->start->reading ; 
  if ($consumption < 0)  
  {       
    $result->error = "Meter appears to have gone backwards ".$consumption ;
    return $result ; 
  }               
  $period = $result->end->enddate - $result->start->enddate ; 

  if ($period == 0)  
  {       
    $result->error = "No suitable meter readings" ;
    return $result ; 
  }
  
  if ($period < 0)  
  {       
    $result->error = "Time appears to have gone backwards" ;
    return $result ; 
  }
  // we now have the closest meter readings but we now need to 
  // calculate an adjustment which will scale the value read 
  // between meter readings into the period for which we are 
  // about to report. 

  $adjust_for_period = ($end - $start)/$period ;
  
  // create an artificial 'one off', scope 0 carbon stamp 
  
  $node = new Result();  // Hm wrong object
  $node->sourceclass = $time_sorted_ar[0]->sourceclass ; 
  $node->adjustment = $time_sorted_ar[0]->adjustment ; 
  $node->tokwh0 = $time_sorted_ar[0]->tokwh0 ; 
  $node->tokwh1 = $time_sorted_ar[0]->tokwh1 ; 
  $node->toco2 = $time_sorted_ar[0]->toco2 ;  
  $node->reading = $consumption;  
   
  
  _invoke_calculate($result, $node, $adjust_for_period) ;  

  $result->notes = "Using readings ". 
  $result->start->reading. " on " . substr(format_date($result->start->enddate, "small"),0,10).
   " and ". 
  $result->end->reading. " on " . substr(format_date($result->end->enddate, "small"),0,10).
  "; adjusted by ".sprintf("%.1f", $adjust_for_period); 


  return $result ; 
}

/**
 * daily event
 * The difference betweeen a scope_0 and scope_2 event is that 
 * the scope_2 event repeats the same level of emissions every
 * day and the emissions are not dispersed between start and 
 * date.   
 */


function _scope_2(&$ar, $start, $stop)
{    
  $result = new Result(); 
  _duplicate_check($ar); 

  foreach ($ar as $node)  
  {   
    if (empty($node->startdate))
    {
      if ($node->enddate > $start && $node->enddate <= $stop)
        $begin = $start ; 
      else
        return $result ; 
    }
    else
    {
      // find out the range of dates when this event occurred 
      // and that fall inside the reporting period that we are 
      // interested in. 
      
      $begin = max($node->startdate, $start);
    }
    $end = min($node->enddate, $stop); 

    if ($end > $begin)
    {   
      $adjust_for_repetition = ($end - $begin) / (60*60*24);  
      _invoke_calculate($result, $node, $adjust_for_repetition);
    }  
    return $result ; // process only 1 stamp, ignore duplicates
  }
}

/**
 * We had expected to be passed an array of just 1 carbon_stamp.
 * Log an error if there is more than one. 
 * @param $ar an array of carbon stamps
 */


function _duplicate_check($ar)
{
  if (count($ar) > 1)
  {
    $text = "These identical carbon stamps treated as one:"; 
    foreach ($ar as $stamp)
    {
      $text .= " ".l("node ".$stamp->nid,"/node/".$stamp->nid);
    }    
    drupal_set_message($text, "warning");
  }	
}


/**
 * walk through the array and look for carbon_stamps that are 
 * closest to the required date - used to find meter readings 
 * near the start and end dates of a carbon footprint. 
 *
 * @param $ar array of all carbon_stamps 
 */

function _find_closest($ar, $date)
{ 
  $node_found = null ;
  foreach ($ar as $node)
  {
    $d = $node->enddate - $date ;   
    if ($d < 0) $d = -$d ; 
    if (!isset($diff_found) || $d < $diff_found)
    {
      $diff_found = $d ;  $node_found = $node; 
    }
  }
  return $node_found ;    
}


/**
 * Each carbon source can have a different classname.
 * Invoked from carbon_source.inc
 * 
 * @return a map of class names (keys) and descriptions (values). Descriptions (values) not used yet.
 */

function carbon_calculate_class_map()
{
  return array(
    "simple"        =>  t("simple multiplier"),
    "climatecare" =>    t("model used by commercial offsetting company"),
    "chooseclimate" =>  t("model created by climate mathematician")
    );
}


/**
 * 
 * At the moment the classname is simply mapped to a function name.
 *
 * @param $node a carbon stamp 
 */


function _invoke_calculate(&$result, $node, $adjust)
{ 
  $fn = '_'.$node->sourceclass ; 
  $fn($result, $node, $adjust); 
}
 
/**
 * The simple class calculates emissions with the formula 
 * 
 * emissions = r*m*a 
 * 
 * where r is the value spcified by the user 
 *       m is the multiplier used to convert units of r 
 *       a is the adjustment specified by the user
 *       which might be 48/52 to allow for holidays
 *       from a daily commute.   
 *       (this can be a string and must be evaluated)
 *
 * to the corresponding CO2 value. 
 * 
 * @param $node a carbon stamp 
 */
 
function _simple(&$result, $node, $adjust)
{ 
  $result->units = $node->reading * _evaluate($node->adjustment) * $adjust; 
  $result->co2 =  $result->units * $node->toco2 ; 
  $result->kwh0 = $result->units * $node->tokwh0 ;
  $result->kwh1 = $result->units * $node->tokwh1 ;
}

/**
 *  This formula comes from http://www.chooseclimate.org 
 * 
 
    Fuel per passenger (Data for B-747).
    
    Fuel = [7840 + 10.1 * (distance-250)] (*2 if return)
    (7840 kg take off-climb-descent, 10.1 kg/km cruising.
    Passengers, = 370 * [occupancy] (/1.5 if business) 
    Greenhouse warming:
    CO2 = fuel * (44/12 * 156/184) (molecular masses)

    Total warming effect of CO2,Ozone (made by NOx), water vapour and 
    contrails is about three times greater than effect of CO2 alone
 */


/**
 * $multiplier should only be used to convert distance from some 
 * arbitrary unit to Km. 
 **/

function _chooseclimate(&$result, $node, $adjust)
{
  $model = array(
    "aircraft" => "B747-400",
    "fuel_for_takeoff" => 7840,
    "fuel_cruising_per_km" => 10.1,    
    "climb_distance" => 250,  
    "capacity" => 370,  
    "occupancy" => .80,
    "fuel_conversion" => (44/12 * 156/184),
    "radiative_forcing_index" => 2.7
    );    
  return _flight_calculate($result, $node, $adjust, $model);
}

/**
     Climate care model 

This model is based on the information in a report written 
by the Environmnetal Change Institute for climate care 
and is provided on the climate care website. 
http://www.climatecare.org/_media/documents/pdf/Aviation_Emissions_&_Offsets.pdf

the raw data refrred to on aircraft fuel consumption
http://reports.eea.europa.eu/EMEPCORINAIR3/en/B851vs2.4.pdf

my version of this with average fuel consumption added: 
http://cvs.drupal.org/viewcvs/ checkout /drupal/contributions/modules/carbon/air_fuel_use.html

note the climate care assume 100% occupancy and that on long 
haul flights the 10% of fuel can be be attributed to freight. 
 */

function _climatecare(&$result, $node, $adjust)
{
  $distance_km = $node->reading * $node->toco2 ;

  // switch planes depending on distance 

  $model =  $distance_km < 3500 ? _climatecareB737() : _climatecareB747();  
  
  return _flight_calculate($result, $node, $adjust, $model);
}

function _climatecareB747() {     
  return array(
    "aircraft" => "B747-400",
    "fuel_for_takeoff" => 3402,
    "fuel_cruising_per_km" => 10.1,    // from average kg/mile in EU table
    "climb_distance" => 0,  
    "fuel_conversion" => 3.15,         // http://www.climatecare.org/_media/documents/pdf/Aviation_Emissions_&_Offsets.pdf
    "capacity" => 400,                 // ditto
    "occupancy" => 1.00,               // ditto
    "radiative_forcing_index" => 2.0   // ditto & http://iacweb.ethz.ch/tradeoff/
    );    
}

function _climatecareB737() {   
  return array(
    "aircraft" => "B737-400",
    "fuel_for_takeoff" => 825.4,
    "fuel_cruising_per_km" => 3.01,    // from average kg/mile above
    "climb_distance" => 0,  
    "fuel_conversion" => 3.15,         // http://www.climatecare.org/_media/documents/pdf/Aviation_Emissions_&_Offsets.pdf
    "capacity" => 180,                 // ditto
    "occupancy" => 1.00,               // ditto
    "radiative_forcing_index" => 2.0   // diito & http://iacweb.ethz.ch/tradeoff/
    );    
}

function _climatecareA340() {   
  return array(
    "aircraft" => "Airbus-A340",
    "fuel_for_takeoff" => 2019,
    "fuel_cruising_per_km" => 6.6,    // from average kg/mile in EU table
    "climb_distance" => 0,  
    "fuel_conversion" => 3.15,         // http://www.climatecare.org/_media/documents/pdf/Aviation_Emissions_&_Offsets.pdf
    "capacity" => 295,                 // ditto
    "occupancy" => 1.00,               // ditto
    "radiative_forcing_index" => 2.0   // diito & http://iacweb.ethz.ch/tradeoff/
    );    
}


/**
 * the choose climate model is basically a simple
 * first order polynomial and I have used it for 
 * the climate care model as well. 
 */


function _flight_calculate(&$result, $node, $adjust, $model)
{ 
  $distance = $node->reading * $node->toco2 ;
  
  $flight_takeoff = $model["fuel_for_takeoff"] ;  
    
  if ($distance > $model["climb_distance"]) 
    $flight_cruising += ($model["fuel_cruising_per_km"] * ($distance-$model["climb_distance"]));  // Kg
      
  $fuel = ($flight_takeoff + $flight_cruising)/($model["capacity"] * $model["occupancy"]) ;   //.Kg 
    
  $co2 = $fuel * $model["fuel_conversion"];  // convert fuel to CO2
  
  // the radiative_forcing_index below (expected to be between 
  // 1.8 and around 2.7) includes effect of the non CO2 gasses

  $co2_and_other_gasses = $co2 * $model["radiative_forcing_index"] ;
  
  $result->notes =  sprintf("Assumptions: aircraft is a %s" .
            " which uses %s Kg of fuel on takeoff/landing, ". 
            " and %s Kg/Km while cruising; ".
            " %s%% of %s seats occupied," .
            " radiative forcing index is %s." .
            " Calculation: CO2 per passenger is %2d Kg," . 
            " all gasses (expressed as CO2 equivalent) is %2d Kg.",  
            $model["aircraft"], $model["fuel_for_takeoff"],  $model["fuel_cruising_per_km"], 
            100 * $model["occupancy"],$model["capacity"], $model["radiative_forcing_index"],            
            $co2, $co2_and_other_gasses); 

  unset($result->units) ;  
  $result->co2 = $co2_and_other_gasses * _evaluate($node->adjustment) * $adjust ; // Kg    
}


/** 
 * Create a printable carbon footprint for a given period (usually a year) 
 * based on on all the available carbon_stamps.  
 * 
 * @param $results a complicated structure containing all results of calculations
 *  
 * @return a string containing a table that shows the carbon footprint.
 */

 
function carbon_calculate_present_results(&$results, $footprint)
{
  $rows = array();    
  $grand_total = 0 ;
  $numupdates = 0 ;      
  $notes = array();  
 

  // now go through the results and create a table 
  // containing the results and then add the notes 
  // to the end of the table. 
  
  foreach ($results->items as $result)
  {      
    $result_as_text = null ;
    $column_notes = "" ;
    if (!empty($result->notes))
    {
      $notes[] = $result->notes; 
      $column_notes =  " [".count($notes)."]" ;  
    }       

    if (!empty($result->error))
    {
      $result_as_text = $result->error ; 
    }
    else
    {
      if (!empty($result->co2))  
        $result_as_text = sprintf("%.0f",$result->co2);
    }
    if (!empty($result_as_text))
    {
      $select = $result->select ; 
      $daterange = "" ; 
      if ($select[0]->scope == 1)
      {
        if ($result->start->enddate != $result->end->enddate)
          $daterange = format_date($result->start->enddate, "custom", "Y-m-d")." and " ;
          $daterange .= format_date($result->end->enddate, "custom", "Y-m-d");
        }
        else
        {
          if (!empty($select[0]->startdate) && ($select[0]->startdate != $select[0]->enddate))
            $daterange = format_date($select[0]->startdate, "custom", "Y-m-d")." to ";          
          $daterange .= format_date($select[0]->enddate, "custom", "Y-m-d");
          $result->end = $select[0] ; 
        }
        $units = is_numeric($result->units) ? sprintf("%.0f",$result->units) : $result->units ; 
        $row = array(
          $daterange,
          carbon_link_from_title($select[0],$footprint)." ".$column_notes,
          (isset($units) ? $units." ".$select[0]->units." of " : "").l($select[0]->sourcemodel."/".$select[0]->sourcetitle, 'node/'.$select[0]->sourceid));
        if ($results->enablekwh0)
          $row[] = sprintf("%.0f",$result->kwh0);
        if ($results->enablekwh1)
          $row[] = sprintf("%.0f",$result->kwh1);
        if ($results->enableco2)
          $row[] = sprintf("%.0f",$result_as_text);
        $ops = carbon_colored_link($result->end,"edit", "node/".$result->end->nid."/edit","destination=node/".$footprint->nid, "#080")."&nbsp;&nbsp;".
        carbon_colored_link($result->end,"clone","node/add/carbon_stamp","clone=".$result->end->nid, "#800") ;
        $row[] = $ops ;
        $rows[] = $row ;  
      $numupdates++ ;
    }  
  }
  if (!empty($numupdates))
  {
    $out = "<h1 class='title'>" ;
    $out .= t("Emissions from ");
    $out .= substr(format_date($results->start, "small"),0,10)." to " ;
    $out .= substr(format_date($results->end, "small"),0,10) ;
    $out .= "</h1>" ; 
    $row = array("total", null, null);
     
    if ($results->enablekwh0)
      $row[] = sprintf("%.0f",$results->grand_kwh0); 
      
    if ($results->enablekwh1)
      $row[] = sprintf("%.0f",$results->grand_kwh1); 
      
    if ($results->enableco2)
      $row[] = sprintf("%.0f",$results->grand_co2);
      
    // pass the footprint and see whether we are allowed to update 
    // it rather than asking if we can crate a new stamp as the former 
    // check makes sure we own the existing footprint.
    $row[] = carbon_colored_link($footprint, "add new","node/add/carbon_stamp", "destination=node/".$footprint->nid, "#CC0");
    $rows[] = $row ; 

    $header = array("date","title","source");
    if ($results->enablekwh0)
      $header[] = "renewable KwH" ; 
    if ($results->enablekwh1)
      $header[] = "non renewable KwH" ; 
    if ($results->enableco2)
      $header[] = "Kg of CO2" ;
    $header[] = "Operations";

    $out .= theme('table', $header, $rows);

    $i = 1 ; foreach ($notes as $note)
    {
      $out .= "<p class='link'>[".$i++."] ".$note."</p>" ; 
    }
  }    
  return $out ;     
}



?>
